/*
 * Goldwater Utilities
 * - Command Line Interface
 *
 * Based on readline examples: 
 *     Copyright (C) 1989, 1992 Free Software Foundation, Inc.
 *
 *     This file contains the GNU History Library (the Library), a set of
 *     routines for managing the text of previously typed lines.
 *
 *     The Library is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2, or (at your option)
 *     any later version.
 *
 *     The Library is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     General Public License for more details.
 *
 *     The GNU General Public License is often shipped with GNU software, and
 *     is generally kept in a file called COPYING or LICENSE.  If you do not
 *     have a copy of the license, write to the Free Software Foundation,
 *     59 Temple Place, Suite 330, Boston, MA 02111 USA. 
 *
 * Copyright 1996-2002 netFluid Technology Limited
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * --------------------------------------------------------------------------
 */
#include <sys/types.h>
#ifdef HAVE_SYS_FILE_H
#  include <sys/file.h>
#endif
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>

#if defined (HAVE_STRING_H)
#  include <string.h>
#else /* !HAVE_STRING_H */
#  include <strings.h>
#endif /* !HAVE_STRING_H */

#define WANT_COMMAND_TABLE
#  include "command.h"
#undef  WANT_COMMAND_TABLE

/* Sort out backwards compatibility */
#if RL_READLINE_VERSION >= 0x0402
#  define RL_COMPLETION_MATCHES rl_completion_matches
#else
#  define RL_COMPLETION_MATCHES completion_matches
#endif

extern char *xmalloc ();

char admin_prompt[ PROMPT_MAX + 1];

static char hist_file[256] = {0};
static sigjmp_buf jbuf;

/* Prototypes */
static char  *command_generator    (const char *, int);
static char  *variables_generator ( const char *text, int state );
static char **goldwater_completion (const char *, int, int);
int ci_strncmp( const char *str1, const char *str2, int len);

/* ------------------------------------------------------------------------- */

static void
catch_ctrlC( int sig )
{
  siglongjmp( jbuf, 1 );  /* Dive back to setsigjmp..... */
}

/* ------------------------------------------------------------------------- */

RetValue
get_command_line( CString *lineCS )
{
  struct sigaction sa, sao;
  char *line = NULL;
  char prompt[PROMPT_MAX + 10 ];

  free_cstring( lineCS );

  sprintf( prompt, "%s> ", admin_prompt );

  /* Grab Ctrl-C here, and sigjump out if it's pressed.
   * Throw a blank lines by returning an empty lineCS
   */
  sa.sa_handler = catch_ctrlC;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = SA_RESTART;
  sigaction( SIGINT, &sa, &sao );

  switch( sigsetjmp( jbuf, 1 ) ) {
  case 0:
    line = readline (prompt);

    if(!line) return Failure;
    WrapCStringMallocedMem( *lineCS, line, strlen(line)+1 );

    skip_lead_space( lineCS );
    skip_tail_space( lineCS );

    /* Force termination of data in lineCS for the benifit of non CS aware funcs
     */
    CSCharPtr( lineCS, CSLenPtr(lineCS) ) = '\0';
	break;

  default:
    if( line ) free(line);
	WrapCStringLen( *lineCS, "\0", 0 );
	printf( "\n" ); /* Throw blank line */
  }

  sa.sa_handler = SIG_IGN;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0;
  sigaction( SIGINT, &sa, &sao );


  if( CSLenPtr( lineCS ) ) {
    add_history ( CSStartPtr(lineCS) );
	append_history( 1, hist_file );
  }

  return Success;
}

/* ------------------------------------------------------------------------- */

#if RL_READLINE_VERSION >= 0x0402
static rl_compentry_func_t *
no_matches ( const char *text, int state )
{
  return ((rl_compentry_func_t *)NULL);
}
#else
static Function *
no_matches ( const char *text, int state )
{
  return ((Function *)NULL);
}
#endif

/* ------------------------------------------------------------------------- */
/* Tell the GNU Readline library how to complete.  We want to try to complete
 * on command names if this is the first word in the line, or on filenames
 * if not.
 */
void
init_readline()
{
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "goldwater";

  /* Tell the completer that we want a crack first. */
  rl_attempted_completion_function = goldwater_completion;
#if RL_READLINE_VERSION >= 0x0402
  rl_completion_entry_function     = (rl_compentry_func_t*)no_matches; /* Disable fname competion */
#else
  rl_completion_entry_function     = (Function*)no_matches; /* Disable fname competion */
#endif

  sprintf( hist_file, "%s/.gwhistory", getenv("HOME") );

  if( read_history( hist_file ) != 0 ) {
    /* create history file up front... */
    open( hist_file, O_CREAT|O_TRUNC, 0666 );
  }
  history_truncate_file( hist_file, 200 );
}

/* ------------------------------------------------------------------------- */
/* Attempt to complete on the contents of TEXT.  START and END bound the
 * region of rl_line_buffer that contains the word to complete.  TEXT is
 * the word to complete.  We can use the entire contents of rl_line_buffer
 * in case we want to do some simple parsing.  Return the array of matches,
 * or NULL if there aren't any. 
 */
static char **
goldwater_completion( const char *text, int start, int end )
{
  char **matches = (char **)NULL;
  CString     partCS     = CSInit;

  /* If this word is at the start of the line, then it is a command
   * to complete.  Otherwise if the line begins with 'set' it's a
   * variable name completion.
   */
  if( start == 0 ) matches = RL_COMPLETION_MATCHES ((char*)text, command_generator);
  else if( start >= 4 ) {

    WrapCStringLen( partCS, rl_line_buffer, start );
    skip_lead_space( &partCS );
    skip_tail_space( &partCS );

    if( strcmpCS( &partCS, "set" ) == 0 ) {
      matches = RL_COMPLETION_MATCHES ((char*)text, variables_generator  );
    } 
  } 

  return (matches);
}

/* ------------------------------------------------------------------------- */

static char *
dupstr( char *s )
{
  char *r = NULL;
  int l = strlen(s) + 1;

  r = xmalloc( l );
  memcpy(r, s, l);
  return (r);
}

/* ------------------------------------------------------------------------- */

static char *
variables_generator ( const char *text, int state )
{
  static int  list_index = 0;
  static int  len        = 0;
  char       *name       = NULL;

  /* If this is a new word to complete, initialise now.
   */
  if (!state) {
    list_index = 0;
    len = strlen (text);
  }

  /* Return the next name which partially matches from the variable list. 
   */
  while( (name = (char*)vartable[list_index].t_text) ) {
    list_index++;

    if( ci_strncmp (name, text, len) == 0) {
	  return (dupstr(name)); 
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* ------------------------------------------------------------------------- */
/* Generator function for command completion.  STATE lets us know whether
 * to start from scratch; without any state (i.e. STATE == 0), then we
 * start at the top of the list. 
 */
static char *
command_generator ( const char *text, int state )
{
  static int  list_index = 0;
  static int  len        = 0;
  char       *name       = NULL;

  /* If this is a new word to complete, initialise now.
   */
  if (!state) {
    list_index = 0;
    len = strlen (text);
  }

  /* Return the next name which partially matches from the command list. 
   */
  while( (name = (char*)cmndtab[list_index].t_text) ) {
    list_index++;

    if( ci_strncmp (name, text, len) == 0) return (dupstr(name)); 
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* ------------------------------------------------------------------------- */

int
ci_strncmp( const char *str1, const char *str2, int len)
{
  int res = 0;
  int i   = 0;
  char c1 = 'A';
  char c2 = 'A';

  i = 0;
  while( i < len && c1 != 0 && c2 != 0 ) {
    c1 = str1[i];
    c2 = str2[i];
		
    if(c1 >= 'a' && c1 <= 'z') c1 -= 32;
    if(c2 >= 'a' && c2 <= 'z') c2 -= 32;

    if(c1 < c2) res = MINUSONE;
    else if(c1 > c2) res = 1;

	i++;
  }

  return res;
}

/* ------------------------------------------------------------------------- */
/* end */
