/*
 * SHM Mutex (semaphores etc)
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * ---------------------------------------------------------------------------
 * $Log: mutex.c,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.14  2003/01/04 19:34:11  chris
 * gw_init() is now quite intelligent with respect to reconnecting.
 *
 * Revision 1.13  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.12  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.11  2002/06/18 00:00:29  chris
 * destroy_bbc_mutex was coded as though individual semaphores in the set were
 * being destroyed, not the set as a whole.  This caused IPC_RMID errors.
 * Tidied loggin in gwmsg.c.
 *
 * Revision 1.10  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.9  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.8  2002/05/12 11:17:43  chris
 * Block signals when in critical section of 'request' code (mainly
 * to stop mutex's being held in a claimed state - an issue with Guardian).
 * Problems are unlikely to occur with real servers.
 *
 * Revision 1.7  2002/03/17 18:44:07  chris
 * Handle the destruction of shm by exposing shmid value.
 *
 * Revision 1.6  2002/02/28 00:55:16  chris
 * Various bug fixes.
 *
 * Revision 1.5  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.4  2001/03/26 18:38:50  chris
 * aquire and release of semaphores no longer gets interrupted by signals
 * (they restart).
 *
 * Revision 1.3  2001/03/26 10:43:14  chris
 * New logic to detect that SHM exists and is in use to avoid resetting
 * it. SHM permissions now 0600.
 *
 * Revision 1.2  2001/03/19 16:09:31  chris
 * Added proper logging.
 *
 * Revision 1.1  2001/03/08 13:35:03  chris
 * Initial revision
 * ---------------------------------------------------------------------------
 */
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <phlib/xl_types.h>

#include <mutex.h>
#include <gwlog.h>
#include <gwgbprm.h>

#ifdef _SEM_SEMUN_UNDEFINED
union semun
{
  int                 val;    /* value for SETVAL              */
  struct semid_ds    *buf;    /* buffer for IPC_STAT & IPC_SET */
  unsigned short int *array;  /* array for GETALL & SETALL     */
  struct seminfo     *__buf;  /* buffer for IPC_INFO           */
};
#endif

static int bbc_semid = -1;

#ifdef DEBUG
#  define GW_DEBUG gw_logf
#else
#  define GW_DEBUG if( 0 ) gw_logf
#endif

/* prototypes */

/* ------------------------------------------------------------------------- */
/*
 * Create a set of semaphores, forcing each one to a value of 1 if force_reset
 * is true.
 */
RetValue
create_bbc_mutex( key_t skey, int force_reset )
{
  int newset = 0;

  /*
   * Try and get a sem set, failing if it's already there.
   */
  errno = 0;
  /* Make sure we've not got a semset before we do a get */
  if( bbc_semid == -1 ) {
    if( (bbc_semid = semget( skey, 
			     NUM_SYS_MUTEX,
			     (0666&~_Mask)|IPC_CREAT|IPC_EXCL)) < 0 && errno != EEXIST ) {
    
      gw_logf( LOG_SERIOUS, "semget failure [mask %ld] %ld (error %ld)",_Mask,(long)skey, errno);
      return Failure;
    }
  
    if( errno == EEXIST ) {
      /* Okay, get the sem set without barfing
       */
      if( (bbc_semid = semget( skey, NUM_SYS_MUTEX, (0666&~_Mask) | IPC_CREAT )) < 0 ) {
        gw_logf( LOG_SERIOUS, "semget failure [mask %ld] %ld (error %ld)",_Mask,(long)skey, errno);
        return Failure;
      }
    } else newset = 1; /* SemSet didn't exist.  FORCE initialisation. */
  }

  /* Initialise all Goldwater system semaphores if force_reset is true,
   * or the semaphore set has only just been created.
   */  
  if( force_reset || newset ) {
    if( reset_all_bbc_mutex() == Failure ) return Failure;
  }

  return Success;
}

/* ------------------------------------------------------------------------- */
/* DANGER Will Robinson!
 */
RetValue
reset_all_bbc_mutex()
{
  int sem_ix = 0;
  union semun su;

  su.val = 1;
  while( sem_ix < NUM_SYS_MUTEX ) {
    if( semctl( bbc_semid, sem_ix, SETVAL, su ) < 0 ) {
	  gw_logf( LOG_SERIOUS, "semctl SETVAL failure (error %ld)", errno );
	  return Failure;
    }
    sem_ix++;
  }
  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
attach_bbc_mutex( key_t skey )
{
  if( bbc_semid == -1 ) {
    if( ( bbc_semid = semget( skey, NUM_SYS_MUTEX, (0666&~_Mask) )) < 0) {
      bbc_semid = -1;
      return Failure;
    }
  }
  return Success;
}

/* ------------------------------------------------------------------------- */
/* Resets bbc_semid
 */
void
detach_bbc_mutex()
{
  bbc_semid = -1;
}

/* ------------------------------------------------------------------------- */

RetValue
aquire_bbc_mutex( int sem )
{
  struct sembuf sb[1];

  sb[0].sem_num = sem;
  sb[0].sem_op  = -1;       /* GET semaphore */
  sb[0].sem_flg = SEM_UNDO;

  GW_DEBUG( LOG_INTDBG, "Get Semaphore %ld", sem );

  do {
    errno = 0;
    if( semop( bbc_semid, sb, 1 ) < 0 ) {
      if( errno == EINTR ) { 
	gw_logf( LOG_INTDBG, "Interrupted semop... redoing" );
	continue;
      }

      gw_logf( LOG_SERIOUS, "semop failure (error %ld)", errno );
      return Failure;
    }
  } while( errno == EINTR );

  GW_DEBUG( LOG_INTDBG, "Got Semaphore %ld", sem );

  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
release_bbc_mutex( int sem )
{
  struct sembuf sb[1];

  sb[0].sem_num = sem;
  sb[0].sem_op  = +1;       /* RELEASE semaphore */
  sb[0].sem_flg = SEM_UNDO;

  GW_DEBUG( LOG_INTDBG, "Release Semaphore %ld", sem );

  do {
    errno = 0;
    if( semop( bbc_semid, sb, 1 ) < 0 ) {
      if( errno == EINTR ) {
	    gw_logf( LOG_INTDBG, "Interrupted semop... redoing" );
	    continue;
      }
      
      gw_logf( LOG_SERIOUS, "semop failure (error %ld)", errno );
      return Failure;
    } 
  } while( errno == EINTR );

  GW_DEBUG( LOG_INTDBG, "Released Semaphore %ld", sem );

  return Success;
}

/* ------------------------------------------------------------------------- */

unsigned short
get_bbc_mutex_value( int sem )
{
  short sval = 0;

  if( (sval = semctl( bbc_semid, sem, GETVAL, 0 ) ) < 0 ) {
    gw_logf( LOG_SERIOUS, "semctl GETVAL failure (error %ld)", errno );
    return -1;
  }

  return (unsigned short)sval;
}

/* ------------------------------------------------------------------------- */
/*
 * Used to get stats of allocated semaphores and usage - used by admin tool
 */
RetValue
semaphore_stats( int semid ) 
{
  return Success;
}

/* ------------------------------------------------------------------------- */
/*
 * This destroys the semaphore set!!!!
 */
RetValue
destroy_bbc_mutex( )
{
  union semun su;

  su.val = 1;
  if( semctl( bbc_semid, 0/*ignored*/, IPC_RMID, su ) < 0 ) {
    gw_logf( LOG_SERIOUS, "semctl IPC_RMID failure (error %ld)", errno );
    return Failure;
  }
  return Success;
}

/* ------------------------------------------------------------------------- */
/* end */
