/*
 * Handles Event and Write API calls (queuing and processing).
 * 
 * Multiplexed Network Client/Server Core Engine
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * -----------------------------------------------------------------------
 * $Id: xlapi.c,v 1.2 2003/03/29 12:21:11 csmith Exp $  $Date: 2003/03/29 12:21:11 $
 * -----------------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_logging.h>
#include <phlib/xl_linklst.h>

#include "gwxlapi.h"
#include "logging.h"
#include "sessions.h"
#include "gwneng.h"

#define DEF_WR_BLOCK_SIZE 4096  /* Default write size */
static int32 r_WR_Block_Size = 0;

extern fd_set _EventBits; /* Event Bits-One for each session(1024 at present)*/
extern int _SFD;          /* netcom.c -> Current Session fd  */
extern int _SelLimit;    /* netcom.c -> max fd to select on */

/* For each connection:
 *  Index from fd back to service_list entry giving handler ptrs,
 *   pointer to private workspace
 *   SSL Connection object
 */
static NetworkConnectionData *NetworkCon = NULL;

static int Sig_Received = 0;
static void sig_in_send( int );

/* List of NetServices (Both Listeners and Non-listeners(defered connection))*/
static NetService *nsrv;

/* Write Request Queue */
static CString wrbufCS  [ WRB_MAX ];  /* Write buffer queue                */
static int     wrb_index[ WRB_MAX ];  /* FD index into Write Buffer Queue  */
static int     wrb_pend [ SES_MAX ];  /* No of pending WRs. 1 each sessn   */ 
static int     wrb_reach[ SES_MAX ];  /* No of DEAD slots to clr when done */
static int     wrb_tpend;             /* Total number of writes pending    */


/* Event Queue */
static int32 evb_last_tless_event[ SES_MAX ]; /* Last tless event in Q  */
static LIST *evb_index[ SES_MAX ];  /* FD index into Event Queues       */
static LIST *evb_pool;

/* ---------------------------------------------------------------------- */
/*
 * This _absolutely_ MUST be called before the main processing loop of events
 * etc...
 */
int
api_event_init()
{
  int ix = 0;

  /* Create session index's */
  ix = 0;
  while( ix < SES_MAX ) {
    xl_List_init( &(evb_index[ix]), sizeof( EventMsg ) );
    evb_last_tless_event[ix] = 0;
    ix++;
  }

  /* Create pool of list nodes - Nodes are transfered out of this
   * pool and added to session index lists as required
   */
  xl_List_init( &evb_pool, sizeof( EventMsg ) );
  ix = 0;
  while( ix < EVT_MAX ) {
    xl_List_insert_space( evb_pool, ListStart );
    ix++;
  }

  return 0;
}

/* ---------------------------------------------------------------------- */
/*
 * Returns the number of events sitting in queue
 */
int
xlapi_events_pending()
{
  return EVT_MAX - xl_List_get_size( evb_pool );
}

/* ---------------------------------------------------------------------- */
/*
 * Returns the number of events sitting in queue FAST INLINE VERSION
 */
__inline__ int
api_events_pending()
{
  return EVT_MAX - xl_List_get_size( evb_pool );
}

/* ---------------------------------------------------------------------- */
int
api_event_deinit()
{
  int ix = 0;

  ix = 0;
  while( ix < SES_MAX ) {
    xl_List_end( &(evb_index[ix]) );
    ix++;
  }

  /* Create pool ot list nodes  */
  xl_List_end( &evb_pool );

  return 0;
}

/* ---------------------------------------------------------------------- */
/*
 * Hub Level API - Call to register events
 */
int
xlapi_register_event( Event ev, int ud )
{
  EventMsg *eq;
  LIST     *ses_events;

  if( xl_List_get_size( evb_pool ) == 0 ) {
    ph_logf( 0, "Error Maximum Events exceeded!" );
    return MINUSONE;
  }

  ses_events = evb_index[_SFD];
  /* Set current to first node of pool */
  xl_List_get_first( evb_pool, (uchar**)&eq ); 

  /* Events should be done as a FIFO, taging new event on to the end  of
   * the event queue... However, because we have timed events, which are
   * ordered in the queue, non-timed events (ie time of zero) must come
   * _first_ in the queue.  Thus we can't insert into the queue at ListEnd!
   * - Event sequencing will be incorrect!
   * So we keep a record of the list position of the last non timed event in
   * the queue and insert at this point.
   */
  /* Transfer node from pool into session event index */
  xl_List_transfer_node( ses_events, evb_pool, evb_last_tless_event[_SFD] );
  evb_last_tless_event[_SFD]++;

  /* Get pointer to new node data */
  eq = (EventMsg*)xl_List_get_current_data( ses_events );
  
#ifdef DEBUG 

  printf( "Adding event to queue   (FD:%d)  Event: %ld/%d\n", 
	  _SFD, ev, ud);
#endif
  eq->event     = ev;
  eq->user_data = ud;
  eq->fd        = _SFD;
  eq->time      = 0;    /* Event is done immediately */
  FD_SET( _SFD, &_EventBits );

  return 0;
}

/* ---------------------------------------------------------------------- */
/*
 * Register a timed event
 */
int
xlapi_register_time_event( Event ev, int32 time )
{
  EventMsg *eq;
  LIST     *ses_events;
  int       node_ix = 0;
  int       node_time = currenttime + time;

  if( xl_List_get_size( evb_pool ) == 0 ) {
    ph_logf( 0, "Error Maximum Events exceeded!" );
    return MINUSONE;
  }

  ses_events = evb_index[_SFD];
  /* Set current to first node of pool */
  xl_List_get_first( evb_pool, (uchar**)&eq ); 

  /* Locate desired slot in this sessions event queue */
  if( xl_List_get_first( ses_events, (uchar**)&eq ) == Success ) {
    do {
      if( eq->time > node_time ) break;
      node_ix++;
    } while( xl_List_get_next( ses_events, (uchar**)&eq ) == Success );
  }
  
  /* If we're off the end, then make sure we point to the last node */
  if( node_ix >= xl_List_get_size( ses_events ) ) node_ix = ListEnd;
  
  /* Transfer node from pool into session event index */
  xl_List_transfer_node( ses_events, evb_pool, node_ix );

  /* Get pointer to new node data */
  eq = (EventMsg*)xl_List_get_current_data( ses_events );
  
#ifdef DEBUG
  printf( "Adding time event to queue  (FD:%d)  Event: %d time %d\n", 
	  _SFD, ev, time);
#endif
  eq->event     = EV_TIME;
  eq->user_data = ev;
  eq->fd        = _SFD;
  eq->time      = node_time;
  FD_SET( _SFD, &_EventBits );

  return 0;
}

/* ---------------------------------------------------------------------- */
/*
 * Register an event intended for a specific session.
 * CARE must be taken when using this function that the intended session
 * does exists and has events enabled.
 *
 * DO NOT call this function if you do not know the target session exists
 *
 * [ Use xlapi_get_session_id() to return the session ID of the currently
 *   active sessiion, store this globaly and use this value when required
 *   by with this function ].
 */
int
xlapi_register_target_event( Event ev, int ud, int targ )
{
  int keep = _SFD;
  int ret  = 0;
  _SFD = targ;
  ret = xlapi_register_event( ev, ud );
  _SFD = keep;

  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * Return the session id of the current active session.  Used with
 * xlapi_register_target_event(...)
 */
__inline__ int
xlapi_get_session_id()
{
  return _SFD;
}

/* ---------------------------------------------------------------------- */
/*
 * Given a session id, become that session and return the old session_id
 */
__inline__ int
xlapi_become_session_id( int ns ) 
{ 
  int os = _SFD; 
  _SFD = ns; 
  return os; 
}

/* ---------------------------------------------------------------------- */
/*
 * This call removes the FIRST event matching the supplied criteria from
 * the event queue.
 *
 */
int
xlapi_remove_event( Event ev, int ud, int which )
{
  EventMsg *eq;
  LIST *ses_events;
  int found = 0;
  int ec = 0;

  ses_events = evb_index[_SFD];
  if( (ec = xl_List_get_size( ses_events )) == 0 ) return 0;

  /* Locate desired slot in this sessions event queue */
  if( xl_List_get_first( ses_events, (uchar**)&eq ) == Success ) {
    do {
      if( eq->event == ev && eq->user_data == ud ) {
	found = 1;
	break;
      }
    } while( xl_List_get_next( ses_events, (uchar**)&eq ) == Success );
  }
  
#ifdef DEBUG
  printf("Delete Found event (FD:%d)i Event: %ld/%d\n", _SFD, 
	 eq->event, eq->user_data);
#endif
  
  if( found == 0 ) {
    ph_logf( 0, "Error Del: Required Event not found in queue (%d:%d)!", ev, ud );
    return MINUSONE;
  }
  
  /* Transfer node out of session event queue and back into pool */
  xl_List_transfer_node( evb_pool, ses_events, ListEnd );

  ec--; 
  /* If we removed an immediate event, then decrement end of tless lst count*/
  if( eq->event != EV_TIME ) evb_last_tless_event[_SFD]--;
  if( ec == 0 ) FD_CLR( _SFD, &_EventBits );
  
  return ec;
}

/* ---------------------------------------------------------------------- */

int
xlapi_total_writes_pending()
{
  return wrb_tpend;
}

/* ---------------------------------------------------------------------- */

int
xlapi_writes_pending()
{
  return wrb_pend[_SFD];
}

/* ---------------------------------------------------------------------- */
/*
 * Function takes a CString of data and sends it out in blocks.
 *
 * It is extremely important that the data remain static until it has
 * been written out completely.  That is, until the invoking session
 * receives a EV_WRDONE event.
 *
 * Returns 1 if we've not hit the limit of WR_PER_SES pending writes per
 * session (this is only a soft limit. We guarantee this many per session,
 * though going over this is safe in a lightly loaded system).
 *
 * Returns 0 if we hit or exceed WR_PER_SES pending writes per session.
 *
 * Returns -1 if all write slots across all sessions have been exhausted.
 */
int
xlapi_write( CString *data )
{
  register int six = _SFD * WR_PER_SES; /* initial bucket */
  register int  ix = 0;

  ix = six;
  /* Find next available slot, starting at hashpoint */
  while( wrb_index[ix] != 0 && ix < WRB_MAX ) ix++; 

  if( ix == WRB_MAX ) {
    /* Hit end of queue, so loop back to the start */
    ix = 0;
    while( wrb_index[ix] != 0 && ix < six ) ix++; 
    if( ix == six ) {
      ph_logf( 0, "Error Maximum Writebuffers exceeded!" );
      return MINUSONE;
    }
  }
#ifdef DEBUG
  printf( "Adding data to write slot %d   (FD:%d)\n", ix, _SFD );  
#endif  
  /* Claim index slot */
  wrb_index[ix] = _SFD;
  wrb_reach[_SFD]++;

  /* First time in then set wrfd */
  if( wrb_pend[_SFD]++ == 0 ) _netcom_set_wrfd( _SFD );
  
  wrb_tpend++; /* Total writes pending counter */
  
  /* Insert buffer details into table */
  MirrorCString( wrbufCS[ ix ], (*data) );

  return (WR_PER_SES - wrb_pend[_SFD]) > 0 ? 1 : 0;
}

/* ---------------------------------------------------------------------- */
/*
 * Function takes a uchar ptr to data and a length, and sends it out in blocks.
 *
 * It is extremely important that the data remain static until it has
 * been written out completely.  That is, until the invoking session
 * receives a EV_WRDONE event.
 *
 * Returns 1 if we've not hit the limit of WR_PER_SES pending writes per
 * session (this is only a soft limit. We guarantee this many per session,
 * though going over this is safe in a lightly loaded system).
 *
 * Returns 0 if we hit or exceed WR_PER_SES pending writes for this session.
 *
 * Returns -1 if all write slots across all sessions have been exhausted.
 */

int
xlapi_write_data( uchar *data, int len )
{
  register int six = _SFD * WR_PER_SES;
  register int  ix = 0;
 
  ix = six;
  /* Find next available slot, starting at hashpoint */
  while( wrb_index[ix] != 0 && ix < WRB_MAX ) ix++; 

  if( ix == WRB_MAX ) {
    ix = 0; /* We hit the end, so loop back round.  Circular Queue thing */
    while( wrb_index[ix] != 0 && ix < six ) ix++; 
    if( ix == six ) {
      ph_logf( 0, "Error Maximum Writebuffers exceeded!" );
      return MINUSONE;
    }
  }

  /* Claim index slot */
  wrb_index[ix] = _SFD;
  wrb_reach[_SFD]++;

  /* First time in then set wrfd */
  if( wrb_pend[_SFD]++ == 0 ) _netcom_set_wrfd( _SFD );

  wrb_tpend++; /* Total writes pending counter */

  /* Insert buffer details into table */
  WrapCStringLen( wrbufCS[ ix ], data, len );

  return (WR_PER_SES - wrb_pend[_SFD]) > 0 ? 1 : 0;
}

/* ---------------------------------------------------------------------- */

int
maintain_write_buffer( int fd ) 
{
  register int l, ll, ix;
  int six;

  six = ix = fd * WR_PER_SES;
  /* Find appropriate output buffer */
  while( wrb_index[ix] != fd && ix < WRB_MAX ) ix++;
  if( ix == WRB_MAX ) {
    ix = 0;
    while( wrb_index[ix] != fd && ix < six ) ix++;
    if( ix == six ) {
      ph_logf( 0, "Error! Required Writebufer not found for output!" );
      return MINUSONE;
    }
  }
  
  if( (l  = CSLen( wrbufCS[ ix ] )) > 0 ) {
    
    ll = (l>r_WR_Block_Size) ? r_WR_Block_Size : l;
    
#if 0
    printf( "Sending fd:%d (%d,%d) (%d bytes) | ", 
	    fd, 
	    *( CSStart(wrbufCS[ix]) + 0 ), 
	    *( CSStart(wrbufCS[ix]) + 1 ), 
	    ll );
#endif
    Sig_Received = 0;
    l = send( fd, (char*)CSStart( wrbufCS[ ix ] ), ll, MSG_DONTWAIT );
    
  } else ph_logf(  0, 
		  "maintain_write_buffer: CString is empty - fd: %d", fd );
  
  if( Sig_Received ) {
    Sig_Received = 0;
    /* Bollocks */
    ph_logf(  0, "Signal received = %d - Drop Connection", Sig_Received );
    
    return (-1); /* Drop Connection */
  }
  
  if( l  < 0 ) {
    if( errno != EWOULDBLOCK ) {
      ph_logf(  0, "Error sending data to fd: %d", fd );
    } else l = 0;
  } else {
    CSLen  ( wrbufCS[ix] ) -= l;
    CSStart( wrbufCS[ix] ) += l;
  }
  
#if 0
  printf( "Bytes remaining... %d\n", CSLen(wrbufCS[ix]) );
#endif
  if( CSLen( wrbufCS[ix] ) <= 0 ) {
    /* Mark index slot as done, but unusable (yet) */
    
    wrb_index[ ix ] = fd + DEAD_SLOT;
    /* Decrement internal count of pending writes */
    wrb_pend[fd]--;
    wrb_tpend--; /* Total writes pending counter */
    
    /* If we've no more pending, then clean up dead slots */
    if( wrb_pend[fd] == 0 ) {
      ix = fd;
      while( wrb_reach[fd] > 0 ) {
	
	if( wrb_index[ix] == (fd + DEAD_SLOT) ) {
	  wrb_index[ix] = 0;
	  wrb_reach[fd]--;
	}
	ix++;
	if( ix == WRB_MAX ) ix = 0; /* Wrap around */
      }
      
      _netcom_reset_wrfd( fd ); /* Reset WR bits for select */
      
      /* Register event saying that all pending writes are done */
      xlapi_register_event( EV_WRDONE, 0 );
    }
  }
  return l;
}

/* ---------------------------------------------------------------------- */

void
api_write_init( int block_size )
{
  register int ix   = 0;

  for( ix = 0; ix < WRB_MAX; ix++ ) {
    wrb_index[ix] = 0;
    if( ix < SES_MAX ) {
      wrb_pend[ix] = wrb_reach[ix] = 0;
    }
  }

  r_WR_Block_Size = block_size;
  wrb_tpend       = 0;

  /* Register signals for writing! 
   * - We don't need to do a sigaction here because we're only doing one
   *   write at a time and so we won't get multiple sigpipe's
   */
  signal( SIGPIPE, sig_in_send );
}

/* ---------------------------------------------------------------------- */

int
xlapi_write_done()
{
  return (wrb_pend[_SFD] != 0);
}

/* ---------------------------------------------------------------------- */
/*
 * Called by event handler on the way out of Modules Event code by HUB
 * to remove event from queue.
 *
 * Returns the number of events pending for this session.
 * *eq points to event descriptor or NULL if there isn't an event to process
 *
 * NOTE: A second call to this function will overwrite the data returned
 *       by a previous call!
 */
int
maintain_event_queue( EventMsg **eq, int fd ) 
{
  static EventMsg eqs;
  EventMsg *ptr;
  LIST *ses_events;
  int ec = 0;
  
  ses_events = evb_index[fd];
  *eq = NULL;
  if( (ec = xl_List_get_size( ses_events )) == 0 ) {
    printf( "No events to process. Hmm This is not correct!\n" );
    return 0;
  }
  
  /* Get next event from list */
  xl_List_get_first( ses_events, (uchar**)&ptr );
  
  /* Now check times.  If we're not ready to process yet then return */
  if( ptr->event == EV_TIME && ptr->time > currenttime ) return ec;

  memcpy( &eqs, ptr, sizeof( EventMsg ) );
  *eq = &eqs;
  
#ifdef DEBUG 
  printf("Found event (FD:%d) Event: %ld/%d\n", fd, 
	 eqs.event, eqs.user_data);
#endif

  /* Transfer node out of session event queue and back into pool */
  xl_List_transfer_node( evb_pool, ses_events, ListEnd );

  ec--; 
  if( ptr->event != EV_TIME ) evb_last_tless_event[fd]--;
  if( ec == 0 ) FD_CLR( fd, &_EventBits );
  
  return ec;
}

/* ---------------------------------------------------------------------- */

void
flush_events( int fd )
{
  EventMsg *ptr;
  LIST *ses_events;
  
  ses_events = evb_index[fd];
  if( xl_List_get_size( ses_events ) == 0 ) return;
  
  /* Take all events from session event queue and return them to the pool */
  do {
    /* Keep pointing at the first node in the session queue */
    xl_List_get_first( ses_events, (uchar**)&ptr );
    
    /* Transfer node out of session event queue and back into pool */
    xl_List_transfer_node( evb_pool, ses_events, ListEnd );
    
  } while( xl_List_get_next( ses_events, (uchar**)&ptr ) == Success );
  
  evb_last_tless_event[fd] = 0;

  /* Clear event's to process mask */
  FD_CLR( fd, &_EventBits );
}


/* ---------------------------------------------------------------------- */

/* THIS DOES NOT CLEAR THE [p]wrfds bits !!!! */
void
flush_writes( int fd )
{
  register int ix;
  
  wrb_tpend -= wrb_pend[fd]; /* Reduce total pending count by ammount */

  wrb_pend[fd] = 0;
  
  ix = fd * WR_PER_SES;  /* Initial Bucket */
  while( wrb_reach[fd] > 0 ) {
    
    if( (wrb_index[ix] & (DEAD_SLOT - 1)) == fd ) {
#ifdef DEBUG
      printf( "flushing write slot %d (%d)\n", ix, fd );
#endif
      wrb_index[ix] = 0;
      wrb_reach[fd]--;
    }
    ix++;
    if( ix >= WRB_MAX ) ix = 0;
  }
}

/* ---------------------------------------------------------------------- */

static void
sig_in_send( int sig )
{
  Sig_Received = sig;
  signal( SIGPIPE, sig_in_send );
}

/* ---------------------------------------------------------------------- */
/* Copy pointer to list of defered connection services
 * The data pointed to by sl must remain constant!
 *
 * Entry: List of defered services
 *        Pointer to Network Connection Data array
 */
void
api_defcon_init( NetService *sl, NetworkConnectionData *cc )
{
  nsrv = sl;
  NetworkCon = cc;
}

/* ---------------------------------------------------------------------- */
/*
 * This function creates a connection session for a given connected fd
 * and service identifier.
 *
 * This is used where services _manually_ make outgoing connections to
 * remote systems and wish to assign the fd obtained to a service to handle
 * the communication.
 *
 * It will be usual to register these fd's with Phoenix Network Services
 * that _do not_ specify a connection port and thus will not listen for
 * incoming connections.  The only way for these services to communicate
 * with remote systems is via this service call.
 *
 * Entry: fd    : The open connection to register with a service
 *        pws   : Pointer to private data buffer within calling session's
 *                service used for private data passing between the two
 *                different by now connected services.
 *        svc_id: The service ID of the required service
 *                ( obtained from xlapi_get_serviceID( ..servicename.. ) )
 *
 */

int
xlapi_register_connection( int session, void *pws, ServiceID service_id )
{
  /* Perform some arbitary validation on service ID */
  if( service_id == -1 ) return -1;
  if( (service_id >> 16) != (service_id & 0xffff) ) return -1;

  NetworkCon[ session ].ServiceIndex = service_id & 0xffff; /* Create an index map */
  NetworkCon[ session ].PrivatePtr   = pws; 
  
  /* Setup an initial event - New Session */
  xlapi_register_target_event( EV_NEW_SESSION, 0, session );

#ifdef DEBUG      
  ph_logf(  9, "## Register remote connection ( fd %d ) ##", fd); 
#endif

  /* _SelLimit is exported from netcom.c - a bit nasty but not much can be
   * done....
   */
  if( session >= _SelLimit ) _SelLimit = session + 1; 
  return session;
}

/* ---------------------------------------------------------------------- */
/*
 * Search list of defered connection services and find one that has a
 * service id string matchin 'svcname'
 *
 * Return a handle ID for this service so that xlapi_register_connection
 * may register an open fd with the correct service.
 *
 * Returns -1 if the service could not be found.
 *
 * Note searching is CASE SENSITIVE !
 */
ServiceID
xlapi_get_serviceID( const char *svcname )
{
  int s = 0;

  /* 'port' is not user in this context, except that a zero value means we've
   * no more services in nsrv.
   */
  while( nsrv[s].port != NETSVCEND && strcmp( nsrv[s].svcidstr, svcname ) != 0 ) s++; 
  if( nsrv[s].port != 0 ) s = -1;
  else 
    if( s > 0xffff ) s = -1;
    else
      /* Add a sort of checksum to the service ID being returned 
       * ie top 16bits = bottom 16bits 
       */
      s |= (s << 16);
  
  return( (ServiceID)s );
}

/* ---------------------------------------------------------------------- */
/*
 * Returns the private workspace pointer for the given session
 */
void *
xlapi_get_private_workspace()
{
	return NetworkCon[ _SFD ].PrivatePtr;
}

/* ---------------------------------------------------------------------- */
/*
 * Returns the ServiceID associated with a particular session ID
 */
const char *
xlapi_get_service_id( int session )
{
  int ix = -1;

  ix = NetworkCon[ session ].ServiceIndex;

  if( ix < 0 ) return NULL;

  return nsrv[ix].svcidstr;
}

/* ---------------------------------------------------------------------- */
/* end */






